"""
Tests for LivingAPI.

This tests the Python and Java implementations of the LivingAPI as well as
direct access via the gateway.

To run the tests, :mod:`pytest` is required. For rerunning flaky tests the
package ``pytest-rerunfailures`` is used.
"""

from conftest import *


###
### Tests
###

def test_user(handler):
	"""
	Check the logged in user.
	"""
	u = user()

	# Check that the logged in user is the user we've used to log in
	vt = handler.make_viewtemplate(
		identifier="livingapi_user_email",
		source="<?print globals.user.email?>",
	)

	assert u == handler.renders(person_app_id, template=vt.identifier)

	# Check that the account name is part of the user ``repr`` output
	vt = handler.make_viewtemplate(
		identifier="livingapi_user_repr",
		source="<?print repr(globals.user)?>",
	)

	assert f" email='{u}'" in handler.renders(person_app_id, template=vt.identifier)


def test_global_hostname(handler):
	"""
	Check that ``globals.hostname`` works.
	"""

	# Check that ``globals.hostname`` is the host we're talking to.
	vt = handler.make_viewtemplate(
		identifier="livingapi_global_hostname",
		source="""
			<?whitespace strip?>
			<?print repr(globals.hostname)?>
		""",
	)

	output = handler.renders(person_app_id, template=vt.identifier)
	assert repr(hostname()) == output


def test_global_datasources(handler, config_apps):
	"""
	Check that ``globals.datasources`` works.
	"""
	c = config_apps

	# Check that the logged in user is the user we've used to log in
	vt = handler.make_viewtemplate(
		la.DataSource(
			identifier="persons",
			app=c.apps.persons,
		),
		la.DataSource(
			identifier="fieldsofactivity",
			app=c.apps.fields,
		),
		identifier="livingapi_global_datasources",
		source="""
			<?whitespace strip?>
			<?print len(globals.datasources)?>
			;
			<?print "persons" in globals.datasources?>
			;
			<?print "fieldsofactivity" in globals.datasources?>
			;
			<?print globals.d_persons.app.id?>
			;
			<?print globals.d_fieldsofactivity.app.id?>
		""",
	)

	output = handler.renders(person_app_id, template=vt.identifier)
	assert f"2;True;True;{person_app_id};{fields_app_id}" == output


def test_app_attributes(handler):
	"""
	Check that ``app`` is the correct one.
	"""
	vt = handler.make_viewtemplate(
		identifier="livingapi_app_attributes",
		source="<?print app.id?>;<?print app.name?>",
	)
	assert f"{person_app_id};LA-Demo: Persons" == handler.renders(person_app_id, template=vt.identifier)


def test_datasources(handler, config_apps):
	"""
	Check that the datasources have the identifiers we expect.
	"""

	c = config_apps

	vt = handler.make_viewtemplate(
		la.DataSource(
			identifier="persons",
			app=c.apps.persons,
		),
		la.DataSource(
			identifier="fieldsofactivity",
			app=c.apps.fields,
		),
		identifier="livingapi_datasources",
		source="<?print ';'.join(sorted(datasources))?>",
	)
	assert "fieldsofactivity;persons" == handler.renders(person_app_id, template=vt.identifier)


def test_output_all_records(handler, config_persons):
	"""
	Simply output all records from all datasources.

	This checks that we don't get any exceptions.
	"""

	c = config_persons

	source = """
		<?for ds in datasources.values()?>
			Datasource/ID: <?print ds.identifier?>
			<?if ds.app is not None and ds.app.records is not None?>
				Datasource/App: <?print ds.app?>
				<?for r in ds.app.records.values()?>
					Record/ID: <?print r.id?>
					Record/Created at: <?print r.createdat?>
					Record/Created by: <?print r.createdby?>
					Record/Updated at: <?print r.updatedat?>
					Record/Updated by: <?print r.updatedby?>
					Record/Update count: <?print r.updatecount?>
					<?for f in r.fields.values()?>
						Record/<?print f.control.identifier?>: <?print repr(f.value)?>
					<?end for?>
				<?end for?>
			<?end if?>
		<?end for?>
	"""

	vt = handler.make_viewtemplate(
		la.DataSource(
			identifier="persons",
			app=c.apps.persons,
		),
		la.DataSource(
			identifier="fieldsofactivity",
			app=c.apps.fields,
		),
		identifier="livingapi_output_all_records",
		source=source,
	)
	handler.renders(person_app_id, template=vt.identifier)


def test_output_all_controls(handler):
	"""
	Simply output all controls from all apps.

	This checks that we don't get any exceptions.
	"""

	source = """
		<?for ds in datasources.values()?>
			<?if ds.app is not None and ds.app.controls is not None?>
				<?for c in ds.app.controls.values()?>
					<?print repr(c)?>
				<?end for?>
			<?end if?>
		<?end for?>
	"""

	handler.make_viewtemplate(
		identifier="livingapi_output_all_controls",
		source=source,
	)

	handler.renders(person_app_id, template="livingapi_output_all_controls")


def test_detail(handler, config_persons):
	"""
	Simply test that detail templates work.
	"""

	source = """
		<?whitespace strip?>
		<?print record.v_firstname?> <?print record.v_lastname?>
	"""

	vt = handler.make_viewtemplate(
		identifier="livingapi_detail",
		source=source,
	)

	assert "Albert Einstein" == handler.renders(
		person_app_id,
		config_persons.persons.ae.id,
		template=vt.identifier,
	)


def test_sort_default_order_is_newest_first(handler, config_persons):
	"""
	Check the the default sort order is descending by creation date.
	"""

	c = config_persons

	source = """
		<?whitespace strip?>
		<?code lastcreatedat = None?>
		<?for p in datasources.persons.app.records.values()?>
			<?if lastcreatedat is not None and lastcreatedat > p.createdat?>
				Bad: <?print lastcreatedat?> > <?print p.createdat?>
			<?end if?>
		<?end for?>
	"""

	vt = handler.make_viewtemplate(
		la.DataSource(
			identifier="persons",
			app=c.apps.persons,
			includecontrols=0,
		),
		identifier="livingapi_sort_default_order_is_newest_first",
		source=source,
	)

	assert not handler.renders(person_app_id, template=vt.identifier)


def test_record_shortcutattributes(handler, config_persons):
	"""
	Find. "Albert Einstein" and output one of his fields in multiple ways
	"""
	c = config_persons

	source = """
		<?whitespace strip?>
		<?code papp = datasources.persons.app?>
		<?code ae = first(r for r in papp.records.values() if r.v_lastname == "Einstein")?>
		<?print repr(ae.fields.firstname.value)?>;
		<?print repr(ae.f_firstname.value)?>;
		<?print repr(ae.values.firstname)?>;
		<?print repr(ae.v_firstname)?>
	"""

	handler.make_viewtemplate(
		la.DataSource(
			identifier="persons",
			app=c.apps.persons,
		),
		identifier="livingapi_record_shortcutattributes",
		source=source,
	)
	assert "'Albert';'Albert';'Albert';'Albert'" == handler.renders(person_app_id, template="livingapi_record_shortcutattributes")


def test_app_shortcutattributes(handler):
	"""
	Access a control and output its fields with in two ways.
	"""
	source = """
		<?whitespace strip?>
		<?print repr(app.controls.firstname.identifier)?>;
		<?print repr(app.c_firstname.identifier)?>
	"""

	vt = handler.make_viewtemplate(
		identifier="livingapi_app_shortcutattributes",
		source=source,
	)
	assert "'firstname';'firstname'" == handler.renders(person_app_id, template=vt.identifier)


def test_insert_record(handler, config_apps):
	"""
	Insert a record into the person app.
	"""
	c = config_apps

	source = """
		<?whitespace strip?>
		<?code r = app.insert(firstname="Isaac", lastname="Newton")?>
		<?print repr(r.v_firstname)?> <?print repr(r.v_lastname)?>;
		<?print r.id?>
	"""

	vt = handler.make_viewtemplate(
		identifier="livingapi_insert_record",
		source=source,
	)
	(output, id) = handler.renders(person_app_id, template=vt.identifier).split(";")

	assert "'Isaac' 'Newton'" == output

	source = f"""
		<?whitespace strip?>
		<?code r = app.records['{id}']?>
		<?print repr(r.v_firstname)?> <?print repr(r.v_lastname)?>
	"""

	vt = handler.make_viewtemplate(
		la.DataSource(
			identifier="persons",
			app=c.apps.persons,
		),
		identifier="livingapi_insert_record_check_result",
		source=source,
	)
	assert "'Isaac' 'Newton'" == handler.renders(person_app_id, template=vt.identifier)


def test_attributes_unsaved_record(handler):
	# Check that ``id``, ``createdat`` and ``createdby`` will be set when the
	# new record is saved
	source = """
		<?whitespace strip?>
		<?code r = app(firstname="Isaac", lastname="Newton")?>
		<?print r.id is None?> <?print r.createdat is None?> <?print r.createdby is None?>
		<?code r.save()?>;
		<?print r.id is None?> <?print r.createdat is None?> <?print r.createdby.email?>
	"""

	vt = handler.make_viewtemplate(
		identifier="livingapi_attributes_unsaved_record_create",
		source=source,
	)
	assert f"True True True;False False {user()}" == handler.renders(person_app_id, template=vt.identifier)

	# Check that ``updatedat`` and ``updatedby`` will be set when the
	# record is saved (this even happens when the record hasn't been changed
	# however in this case no value fields will be changed)
	source = """
		<?whitespace strip?>
		<?code r = app(firstname="Isaac", lastname="Newton")?>
		<?print r.updatedat is None?> <?print r.updatedby is None?>
		<?code r.save()?>;
		<?print r.updatedat is None?> <?print r.updatedby is None?>
		<?code r.save()?>;
		<?print r.updatedat is None?> <?print r.updatedby.email?>
		<?code r.v_date_of_birth = @(1642-12-25)?>
		<?code r.save()?>;
		<?print r.updatedat is None?> <?print r.updatedby.email?>
	"""

	vt = handler.make_viewtemplate(
		identifier="livingapi_attributes_unsaved_record_update",
		source=source,
	)

	assert f"True True;True True;False {user()};False {user()}" == handler.renders(person_app_id, template=vt.identifier)


def test_no_appparams(handler):
	source = "<?print repr(app.params)?>"

	vt = handler.make_viewtemplate(
		identifier="livingapi_no_appparams",
		source=source,
	)

	assert "None" == handler.renders(person_app_id, template=vt.identifier)


def test_appparam_bool(handler, config_apps):
	c = config_apps

	source = """
		<?whitespace strip?>
		<?print repr(app.params.bool_none.value)?>
		;<?print app.params.bool_none.description?>
		;<?print repr(app.params.bool_false.value)?>
		;<?print app.params.bool_false.description?>
		;<?print repr(app.params.bool_true.value)?>
		;<?print app.params.bool_true.description?>
	"""

	vt = handler.make_viewtemplate(
		la.DataSource(
			identifier="persons",
			app=c.apps.persons,
			includeparams=True,
		),
		identifier="livingapi_appparam_bool",
		source=source,
	)

	output = handler.renders(person_app_id, template=vt.identifier)
	assert "None;desc bool_none;False;desc bool_false;True;desc bool_true" == output


def test_appparam_int(handler, config_apps):
	c = config_apps

	source = """
		<?whitespace strip?>
		<?print repr(app.params.int_none.value)?>
		;<?print app.params.int_none.description?>
		;<?print repr(app.params.int_value.value)?>
		;<?print app.params.int_value.description?>
	"""

	vt = handler.make_viewtemplate(
		la.DataSource(
			identifier="persons",
			app=c.apps.persons,
			includeparams=True,
		),
		identifier="livingapi_appparam_int",
		source=source,
	)

	output = handler.renders(person_app_id, template=vt.identifier)
	assert "None;desc int_none;1777;desc int_value" == output


def tcest_livingapi_appparam_number(handler, config_apps):
	c = config_apps

	source = """
		<?whitespace strip?>
		<?print repr(app.params.number_none.value)?>
		;<?print app.params.number_none.description?>
		;<?print repr(app.params.number_value.value)?>
		;<?print app.params.number_value.description?>
	"""

	vt = handler.make_viewtemplate(
		la.DataSource(
			identifier="persons",
			app=c.apps.persons,
			includeparams=True,
		),
		identifier="livingapi_appparam_number",
		source=source,
	)

	output = handler.renders(person_app_id, template=vt.identifier)
	assert "None;desc number_none;42.5;desc number_value" == output


def test_appparam_str(handler, config_apps):
	c = config_apps

	source = """
		<?whitespace strip?>
		<?print repr(app.params.str_none.value)?>
		;<?print app.params.str_none.description?>
		;<?print repr(app.params.str_value.value)?>
		;<?print app.params.str_value.description?>
	"""

	vt = handler.make_viewtemplate(
		la.DataSource(
			identifier="persons",
			app=c.apps.persons,
			includeparams=True,
		),
		identifier="livingapi_appparam_str",
		source=source,
	)

	output = handler.renders(person_app_id, template=vt.identifier)

	assert "None;desc str_none;'gurk';desc str_value" == output


def test_appparam_color(handler, config_apps):
	c = config_apps

	source = """
		<?whitespace strip?>
		<?print repr(app.params.color_none.value)?>
		;<?print app.params.color_none.description?>
		;<?print repr(app.params.color_value.value)?>
		;<?print app.params.color_value.description?>
	"""

	vt = handler.make_viewtemplate(
		la.DataSource(
			identifier="persons",
			app=c.apps.persons,
			includeparams=True,
		),
		identifier="livingapi_appparam_color",
		source=source,
	)

	output = handler.renders(person_app_id, template=vt.identifier)

	assert "None;desc color_none;#369c;desc color_value" == output


def test_appparam_datedelta(handler, config_apps):
	c = config_apps

	source = """
		<?whitespace strip?>
		<?print repr(app.params.datedelta_none.value)?>
		;<?print app.params.datedelta_none.description?>
		;<?print repr(app.params.datedelta_value.value)?>
		;<?print app.params.datedelta_value.description?>
	"""

	vt = handler.make_viewtemplate(
		la.DataSource(
			identifier="persons",
			app=c.apps.persons,
			includeparams=True,
		),
		identifier="livingapi_appparam_datedelta",
		source=source,
	)

	output = handler.renders(person_app_id, template=vt.identifier)
	assert "None;desc datedelta_none;timedelta(days=12);desc datedelta_value" == output


def test_appparam_datetimedelta(handler, config_apps):
	c = config_apps

	source = """
		<?whitespace strip?>
		<?print repr(app.params.datetimedelta_none.value)?>
		;<?print app.params.datetimedelta_none.description?>
		;<?print repr(app.params.datetimedelta_value.value)?>
		;<?print app.params.datetimedelta_value.description?>
	"""

	vt = handler.make_viewtemplate(
		la.DataSource(
			identifier="persons",
			app=c.apps.persons,
			includeparams=True,
		),
		identifier="livingapi_appparam_datetimedelta",
		source=source,
	)

	output = handler.renders(person_app_id, template=vt.identifier)
	assert "None;desc datetimedelta_none;timedelta(days=1, seconds=45296);desc datetimedelta_value" == output


def test_appparam_monthdelta(handler, config_apps):
	c = config_apps

	source = """
		<?whitespace strip?>
		<?print repr(app.params.monthdelta_none.value)?>
		;<?print app.params.monthdelta_none.description?>
		;<?print repr(app.params.monthdelta_value.value)?>
		;<?print app.params.monthdelta_value.description?>
	"""

	vt = handler.make_viewtemplate(
		la.DataSource(
			identifier="persons",
			app=c.apps.persons,
			includeparams=True,
		),
		identifier="livingapi_appparam_monthdelta",
		source=source,
	)

	output = handler.renders(person_app_id, template=vt.identifier)
	assert "None;desc monthdelta_none;monthdelta(3);desc monthdelta_value" == output


def test_appparam_upload(handler, config_apps):
	c = config_apps

	source = """
		<?whitespace strip?>
		<?print repr(app.params.upload_none.value)?>
		;<?print app.params.upload_none.description?>
		;<?print repr(app.params.upload_value.value.mimetype)?>
		;<?print app.params.upload_value.description?>
	"""

	vt = handler.make_viewtemplate(
		la.DataSource(
			identifier="persons",
			app=c.apps.persons,
			includeparams=True,
		),
		identifier="livingapi_appparam_upload",
		source=source,
	)

	output = handler.renders(person_app_id, template=vt.identifier)

	assert "None;desc upload_none;'image/jpeg';desc upload_value" == output


def test_appparam_app(handler, config_apps):
	c = config_apps

	source = """
		<?whitespace strip?>
		<?print repr(app.params.app_none.value)?>
		;<?print app.params.app_none.description?>
		;<?print repr(app.params.app_value.value.id)?>
		;<?print app.params.app_value.description?>
	"""

	vt = handler.make_viewtemplate(
		la.DataSource(
			identifier="persons",
			app=c.apps.persons,
			includeparams=True,
		),
		identifier="livingapi_appparam_app",
		source=source,
	)

	output = handler.renders(person_app_id, template=vt.identifier)

	assert f"None;desc app_none;'{person_app_id}';desc app_value" == output


def test_appparam_otherattributes(handler, config_apps):
	c = config_apps

	source = """
		<?whitespace strip?>
		<?note Use shortcut attribute in all expressions?>
		<?print app.p_str_value.identifier?>
		;<?print app.p_str_value.description?>
		;<?print isdatetime(app.p_str_value.createdat)?>
		;<?print isdefined(app.p_str_value.createdby)?>
		;<?print isdefined(app.p_str_value.updatedat)?>
		;<?print isdefined(app.p_str_value.updatedby)?>
	"""

	vt = handler.make_viewtemplate(
		la.DataSource(
			identifier="persons",
			app=c.apps.persons,
			includeparams=True,
		),
		identifier="livingapi_appparam_otherattributes",
		source=source,
	)

	output = handler.renders(person_app_id, template=vt.identifier)
	assert "str_value;desc str_value;True;True;True;True" == output



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# cython: language_level=3, always_allow_keywords=True

## Copyright 2016-2020 by LivingLogic AG, Bayreuth/Germany
##
## All Rights Reserved

"""
Classes and functions for compiling vSQL expressions.
"""

import datetime, itertools

from ll import color, misc, ul4c, ul4on

try:
	from ll import orasql
except ImportError:
	orasql = None


###
### Helper functions and classes
###

def _offset(pos):
	# Note that we know that for our slices ``start``/``stop`` are never ``None``
	return slice(pos.start-9, pos.stop-9)


def compile(source, vars={}):
	template = ul4c.Template(f"<?return {source}?>")
	expr = template.content[-1].obj
	return AST.fromul4(source, expr, vars)


def compile_and_save(handler, cursor, source, datatype, function, **queryargs):
	if source is None:
		return None
	else:
		args = ", ".join(f"{a}=>:{a}" for a in queryargs)
		query = f"select {function}({args}) from dual"
		cursor.execute(query, **queryargs)
		dump = cursor.fetchone()[0]
		dump = dump.decode("utf-8")
		vars = ul4on.loads(dump)
		ast = compile(source, vars)
		vs_id = ast.save(handler, cursor=cursor)
		cursor.execute(
			"begin vsql_pkg.vsql_validate(:vs_id, :datatype); end;",
			vs_id=vs_id,
			datatype=datatype,
		)
		return vs_id


class Repr:
	def _ll_repr_prefix_(self):
		return f"{self.__class__.__module__}.{self.__class__.__qualname__}"

	def _ll_repr_suffix_(self):
		return f"at {id(self):#x}"

	def __repr__(self):
		parts = itertools.chain(
			(f"<{self._ll_repr_prefix_()}",),
			self._ll_repr_(),
			(f"{self._ll_repr_suffix_()}>",),
		)
		return " ".join(parts)

	def _ll_repr_(self):
		yield from ()

	def _repr_pretty_(self, p, cycle):
		if cycle:
			p.text(f"{self._ll_repr_prefix_()} ... {self._ll_repr_suffix_()}>")
		else:
			with p.group(3, f"<{self._ll_repr_prefix_()}", ">"):
				self._ll_repr_pretty_(p)
				p.breakable()
				p.text(self._ll_repr_suffix_())

	def _ll_repr_pretty_(self, p):
		pass


###
### Core classes
###

class Def(Repr):
	pass


@ul4on.register("de.livinglogic.vsql.field")
class Field(Def):
	def __init__(self, identifier=None, datatype=None, fieldsql=None, joinsql=None, refgroup=None):
		self.identifier = identifier
		self.datatype = datatype
		self.fieldsql = fieldsql
		self.joinsql = joinsql
		self.refgroup = refgroup

	def _ll_repr_(self):
		yield f"{self.identifier!r}"
		yield f"datatype={self.datatype!r}"
		yield f"fieldsql={self.fieldsql!r}"
		if self.joinsql is not None:
			yield f"joinsql={self.joinsql!r}"
		if self.refgroup is not None:
			yield f"refgroup.tablesql={self.refgroup.tablesql!r}"

	def _ll_repr_pretty_(self, p):
		p.text(" ")
		p.pretty(self.identifier)
		p.breakable()
		p.text("datatype=")
		p.pretty(self.datatype)
		p.breakable()
		p.text("fieldsql=")
		p.pretty(self.fieldsql)
		if self.joinsql is not None:
			p.breakable()
			p.text("joinsql=")
			p.pretty(self.joinsql)
		if self.refgroup is not None:
			p.breakable()
			p.text("refgroup.tablesql=")
			p.pretty(self.refgroup.tablesql)

	def ul4ondump(self, encoder):
		encoder.dump(self.identifier)
		encoder.dump(self.datatype)
		encoder.dump(self.fieldsql)
		encoder.dump(self.joinsql)
		encoder.dump(self.refgroup)

	def ul4onload(self, decoder):
		self.identifier = decoder.load()
		self.datatype = decoder.load()
		self.fieldsql = decoder.load()
		self.joinsql = decoder.load()
		self.refgroup = decoder.load()


@ul4on.register("de.livinglogic.vsql.group")
class Group(Def):
	def __init__(self, tablesql=None, **fields):
		self.tablesql = tablesql
		self.fields = fields

	def _ll_repr_(self):
		yield f"tablesql={self.tablesql!r}"
		yield f"with {len(self.fields):,} fields"

	def _ll_repr_pretty_(self, p):
		p.breakable()
		p.text("tablesql=")
		p.pretty(self.tablesql)
		for (fieldname, field) in self.fields.items():
			p.breakable()
			p.text(f"fields.{fieldname}=")
			p.pretty(field)

	def __getitem__(self, key):
		return self.fields[key]

	def ul4ondump(self, encoder):
		encoder.dump(self.tablesql)
		encoder.dump(self.fields)

	def ul4onload(self, decoder):
		self.tablesql = decoder.load()
		self.fields = decoder.load()


class AST(Repr):
	dbnodetype = None
	dbnodevalue = None
	dbdatatype = None

	def __init__(self, source=None, pos=None):
		self._source = source
		self.pos = pos

	@property
	def source(self):
		return self._source[self.pos]

	def dbchildren(self):
		yield from ()

	def save(self, handler, cursor=None, vs_id_super=None, vs_order=None, vss_id=None):
		"""
		Save :obj:`self` to the :class:`DBHandler` :obj:`handler`.

		``cursor``, ``vs_id_super``, ``vs_order`` and ``vss_id`` are used
		internally for recursive calls and should not be passed by the user.
		"""
		if cursor is None:
			cursor = handler.cursor()
		if vss_id is None:
			r = handler.proc_vsqlsource_insert(
				cursor,
				c_user=handler.ide_id,
				p_vss_source=self.source,
			)
			vss_id = r.p_vss_id
		r = handler.proc_vsql_insert(
			cursor,
			c_user=handler.ide_id,
			p_vs_id_super=vs_id_super,
			p_vs_order=vs_order,
			p_vs_nodetype=self.dbnodetype,
			p_vs_value=self.dbnodevalue,
			p_vs_datatype=self.dbdatatype,
			p_vss_id=vss_id,
			p_vs_start=self.pos.start,
			p_vs_stop=self.pos.stop,
		)
		vs_id = r.p_vs_id
		order = 10
		for child in self.dbchildren():
			child.save(handler, cursor, vs_id, order, vss_id)
			order += 10
		return vs_id

	def __str__(self):
		return f"{self.__class__.__name__}: {self.source[self.pos]}"

	def _ll_repr_suffix_(self):
		parts = []
		parts.append("pos=[")
		if self.pos.start is not None:
			parts.append(f"{self.pos.start:,}")
		parts.append(":")
		if self.pos.stop is not None:
			parts.append(f"{self.pos.stop:,}")
		parts.append("] ")
		parts.append(super()._ll_repr_suffix_())
		return "".join(parts)

	@classmethod
	def fromul4(cls, source, node, vars):
		if isinstance(node, ul4c.Const):
			if node.value is None:
				return None_.fromul4(source, node, vars)
			else:
				try:
					vsqltype = _consts[type(node.value)]
				except KeyError:
					raise TypeError(f"constant of type {misc.format_class(node.value)} not supported!") from None
				return vsqltype.fromul4(source, node, vars)
		else:
			try:
				vsqltype = _ul42vsql[type(node)]
			except KeyError:
				pass
			else:
				return vsqltype.fromul4(source, node, vars)
		if isinstance(node, ul4c.Var):
			field = vars.get(node.name, None)
			return FieldRef(source, _offset(node.pos), None, node.name, field)
		elif isinstance(node, ul4c.Attr):
			vsqlnode = cls.fromul4(source, node.obj, vars)
			if isinstance(vsqlnode, FieldRef) and isinstance(vsqlnode.field, Field) and vsqlnode.field.refgroup:
				try:
					field = vsqlnode.field.refgroup.fields[node.attrname]
				except KeyError:
					try:
						field = vsqlnode.field.refgroup.fields["*"]
					except KeyError:
						pass # Fall through to return a generic :class:`Attr` node
					else:
						return FieldRef(source, _offset(node.pos), vsqlnode, node.attrname, field)
				else:
					return FieldRef(source, _offset(node.pos), vsqlnode, node.attrname, field)
			return Attr(source, _offset(node.pos), vsqlnode, node.attrname)
		elif isinstance(node, ul4c.Call):
			vsqlnode = cls.fromul4(source, node.obj, vars)
			args = []
			for arg in node.args:
				if not isinstance(arg, ul4c.PosArg):
					raise TypeError(f"Can't compile UL4 expression of type {misc.format_class(arg)}!")
				args.append(AST.fromul4(source, arg.value, vars))
			if isinstance(vsqlnode, FieldRef):
				if vsqlnode.parent is not None:
					return Meth(source, _offset(node.pos), vsqlnode.parent, vsqlnode.identifier, args)
				else:
					return Func(source, _offset(node.pos), vsqlnode.identifier, args)
			elif isinstance(vsqlnode, Attr):
				return Meth(source, _offset(node.pos), vsqlnode.obj, vsqlnode.attrname, args)
		raise TypeError(f"Can't compile UL4 expression of type {misc.format_class(node)}!")

	def ul4ondump(self, encoder):
		encoder.dump(self._source)
		encoder.dump(self.pos)

	def ul4onload(self, decoder):
		self._source = decoder.load()
		self.pos = decoder.load()


class Const(AST):
	pass


@ul4on.register("de.livinglogic.vsql.none")
class None_(Const):
	dbnodetype = "const_none"
	dbdatatype = "null"

	@classmethod
	def fromul4(cls, source, node, vars):
		return cls(source, _offset(node.pos))


class _ConstWithValue(Const):
	def __init__(self, source=None, pos=None, value=None):
		super().__init__(source, pos)
		self.value = value

	@property
	def dbnodevalue(self):
		return self.value

	def _ll_repr_(self):
		yield f"value={self.value!r}"

	@classmethod
	def fromul4(cls, source, node, vars):
		return cls(source, _offset(node.pos), node.value)

	def _ll_repr_pretty_(self, p):
		p.breakable()
		p.text("value=")
		p.pretty(self.value)

	def ul4ondump(self, encoder):
		super().ul4ondump(encoder)
		encoder.dump(self.value)

	def ul4onload(self, decoder):
		super().ul4onload(decoder)
		self.value = decoder.load()


@ul4on.register("de.livinglogic.vsql.bool")
class Bool(_ConstWithValue):
	dbnodetype = "const_bool"
	dbdatatype = "bool"

	@property
	def dbnodevalue(self):
		return "True" if self.value else "False"


@ul4on.register("de.livinglogic.vsql.int")
class Int(_ConstWithValue):
	dbnodetype = "const_int"
	dbdatatype = "int"

	@property
	def dbnodevalue(self):
		return str(self.value)


@ul4on.register("de.livinglogic.vsql.number")
class Number(_ConstWithValue):
	dbnodetype = "const_number"
	dbdatatype = "number"

	@property
	def dbnodevalue(self):
		return repr(self.value)


@ul4on.register("de.livinglogic.vsql.str")
class Str(_ConstWithValue):
	dbnodetype = "const_str"
	dbdatatype = "str"


@ul4on.register("de.livinglogic.vsql.color")
class Color(_ConstWithValue):
	dbnodetype = "const_color"
	dbdatatype = "color"

	@property
	def dbnodevalue(self):
		c = self.value
		return f"{c.r():02x}{c.g():02x}{c.b():02x}{c.a():02x}"


@ul4on.register("de.livinglogic.vsql.date")
class Date(_ConstWithValue):
	dbnodetype = "const_date"
	dbdatatype = "date"

	@property
	def dbnodevalue(self):
		return f"{self.value:%Y-%m-%d}"


@ul4on.register("de.livinglogic.vsql.datetime")
class DateTime(_ConstWithValue):
	dbnodetype = "const_datetime"
	dbdatatype = "datetime"

	@property
	def dbnodevalue(self):
		return f"{self.value:%Y-%m-%dT%H:%M:%S}"


@ul4on.register("de.livinglogic.vsql.list")
class List(AST):
	dbnodetype = "list"

	def __init__(self, source=None, pos=None):
		super().__init__(source, pos)
		self.items = []

	def _ll_repr_(self):
		yield f"with {len(self.items):,} items"

	def _ll_repr_pretty_(self, p):
		for item in self.items:
			p.breakable()
			p.pretty(item)

	@classmethod
	def fromul4(cls, source, node, vars):
		self = cls(source, _offset(node.pos))
		for item in node.items:
			if not isinstance(item, ul4c.SeqItem):
				raise TypeError(f"Can't compile UL4 expression of type {misc.format_class(item)}!")
			self.items.append(AST.fromul4(source, item.value, vars))
		return self

	def dbchildren(self):
		yield from self.items

	def ul4ondump(self, encoder):
		super().ul4ondump(encoder)
		encoder.dump(self.items)

	def ul4onload(self, decoder):
		super().ul4onload(decoder)
		self.items = decoder.load()


@ul4on.register("de.livinglogic.vsql.set")
class Set(AST):
	dbnodetype = "set"

	def __init__(self, source=None, pos=None):
		super().__init__(source, pos)
		self.items = []

	def _ll_repr_(self):
		yield f"with {len(self.items):,} items"

	def _ll_repr_pretty_(self, p):
		for item in self.items:
			p.breakable()
			p.pretty(item)

	@classmethod
	def fromul4(cls, source, node, vars):
		self = cls(source, _offset(node.pos))
		for item in node.items:
			if not isinstance(item, ul4c.SeqItem):
				raise TypeError(f"Can't compile UL4 expression of type {misc.format_class(item)}!")
			self.items.append(AST.fromul4(source, item.value, vars))
		return self

	def dbchildren(self):
		yield from self.items

	def ul4ondump(self, encoder):
		super().ul4ondump(encoder)
		encoder.dump(self.items)

	def ul4onload(self, decoder):
		super().ul4onload(decoder)
		self.items = decoder.load()


@ul4on.register("de.livinglogic.vsql.fieldref")
class FieldRef(AST):
	dbnodetype = "field"

	def __init__(self, source=None, pos=None, parent=None, identifier=None, field=None):
		"""
		Create a :class:`FieldRef` object.

		There are three possible scenarios with respect to :obj`identifier` and
		:obj:`field`:

		``field is not None and field.identifier == identifier``
			In this case we have a valid :class:`Field` that describes a real
			field.

		``field is not None and field.identifier != identifier and field.identifier == "*"``
			In this case :obj:`field` is the :class:`Field` object for the generic
			typed request parameters. E.g. when the vSQL expression is
			``params.str.foo`` then :obj:`field` references the :class:`Field` for
			``params.str.*``, so ``field.identifier == "*" and
			``identifier == "foo"``.

		``field is None``
			In this case the field is unknown.
		"""
		super().__init__(source, pos)
		self.parent = parent
		# Note that ``identifier`` might be different from ``field.identifier``
		# if ``field.identifier == "*"``.
		self.identifier = identifier
		# Note that ``field`` might be ``None`` when the field can't be found.
		self.field = field

	@property
	def dbdatatype(self):
		return self.field.datatype

	@property
	def dbnodevalue(self):
		identifierpath = []
		node = self
		while node is not None:
			identifierpath.insert(0, node.identifier)
			node = node.parent
		return ".".join(identifierpath)

	def _ll_repr_(self):
		if self.parent is not None:
			yield f"parent={self.parent!r}"
		if self.field is None or self.field.identifier != self.identifier:
			yield f"identifier={self.identifier!r}"
		if self.field is not None:
			yield f"field={self.field!r}"

	def _ll_repr_pretty_(self, p):
		p.text(" ")
		p.pretty(self.name)
		if self.parent is not None:
			p.breakable()
			p.text("parent=")
			p.pretty(self.parent)
		if self.field is None or self.field.identifier != self.identifier:
			p.breakable()
			p.text("identifier=")
			p.pretty(self.identifier)
		if self.field is not None:
			p.breakable()
			p.text("field=")
			p.pretty(self.field)

	def ul4ondump(self, encoder):
		super().ul4ondump(encoder)
		encoder.dump(self.parent)
		encoder.dump(self.identifier)
		encoder.dump(self.field)

	def ul4onload(self, decoder):
		super().ul4onload(decoder)
		self.parent = decoder.load()
		self.identifier = decoder.load()
		self.field = decoder.load()


class Binary(AST):
	def __init__(self, source=None, pos=None, obj1=None, obj2=None):
		super().__init__(source, pos)
		self.obj1 = obj1
		self.obj2 = obj2

	def dbchildren(self):
		yield self.obj1
		yield self.obj2

	def _ll_repr_(self):
		yield f"obj1={self.obj1!r}"
		yield f"obj2={self.obj2!r}"

	def _ll_repr_pretty_(self, p):
		p.breakable()
		p.text("obj1=")
		p.pretty(self.obj1)
		p.breakable()
		p.text("obj2=")
		p.pretty(self.obj2)

	@classmethod
	def fromul4(cls, source, node, vars):
		return cls(source, _offset(node.pos), AST.fromul4(source, node.obj1, vars), AST.fromul4(source, node.obj2, vars))

	def ul4ondump(self, encoder):
		super().ul4ondump(encoder)
		encoder.dump(self.obj1)
		encoder.dump(self.obj2)

	def ul4onload(self, decoder):
		super().ul4onload(decoder)
		self.obj1 = decoder.load()
		self.obj2 = decoder.load()


@ul4on.register("de.livinglogic.vsql.eq")
class EQ(Binary):
	dbnodetype = "cmp_eq"


@ul4on.register("de.livinglogic.vsql.ne")
class NE(Binary):
	dbnodetype = "cmp_ne"


@ul4on.register("de.livinglogic.vsql.lt")
class LT(Binary):
	dbnodetype = "cmp_lt"


@ul4on.register("de.livinglogic.vsql.le")
class LE(Binary):
	dbnodetype = "cmp_le"


@ul4on.register("de.livinglogic.vsql.gt")
class GT(Binary):
	dbnodetype = "cmp_gt"


@ul4on.register("de.livinglogic.vsql.ge")
class GE(Binary):
	dbnodetype = "cmp_ge"


@ul4on.register("de.livinglogic.vsql.add")
class Add(Binary):
	dbnodetype = "binop_add"


@ul4on.register("de.livinglogic.vsql.sub")
class Sub(Binary):
	dbnodetype = "binop_sub"


@ul4on.register("de.livinglogic.vsql.mul")
class Mul(Binary):
	dbnodetype = "binop_mul"


@ul4on.register("de.livinglogic.vsql.floordiv")
class FloorDiv(Binary):
	dbnodetype = "binop_floordiv"


@ul4on.register("de.livinglogic.vsql.truediv")
class TrueDiv(Binary):
	dbnodetype = "binop_truediv"


@ul4on.register("de.livinglogic.vsql.mod")
class Mod(Binary):
	dbnodetype = "binop_mod"


@ul4on.register("de.livinglogic.vsql.and")
class And(Binary):
	dbnodetype = "binop_and"


@ul4on.register("de.livinglogic.vsql.or")
class Or(Binary):
	dbnodetype = "binop_or"


@ul4on.register("de.livinglogic.vsql.contains")
class Contains(Binary):
	dbnodetype = "binop_contains"


@ul4on.register("de.livinglogic.vsql.notcontains")
class NotContains(Binary):
	dbnodetype = "binop_notcontains"


@ul4on.register("de.livinglogic.vsql.is")
class Is(Binary):
	dbnodetype = "binop_is"


@ul4on.register("de.livinglogic.vsql.isnot")
class IsNot(Binary):
	dbnodetype = "binop_isnot"


@ul4on.register("de.livinglogic.vsql.item")
class Item(Binary):
	dbnodetype = "binop_item"

	@classmethod
	def fromul4(cls, source, node, vars):
		if isinstance(node.obj2, ul4c.Slice):
			return Slice.fromul4(source, node, vars)
		return super().fromul4(source, node, vars)


@ul4on.register("de.livinglogic.vsql.shiftleft")
class ShiftLeft(Binary):
	dbnodetype = "binop_shiftleft"


@ul4on.register("de.livinglogic.vsql.shiftright")
class ShiftRight(Binary):
	dbnodetype = "binop_shiftright"


@ul4on.register("de.livinglogic.vsql.bitand")
class BitAnd(Binary):
	dbnodetype = "binop_bitand"


@ul4on.register("de.livinglogic.vsql.bitor")
class BitOr(Binary):
	dbnodetype = "binop_bitor"


@ul4on.register("de.livinglogic.vsql.bitxor")
class BitXOr(Binary):
	dbnodetype = "binop_bitxor"


class Unary(AST):
	def __init__(self, source=None, pos=None, obj=None):
		super().__init__(source, pos)
		self.obj = obj

	def dbchildren(self):
		yield self.obj

	def _ll_repr_(self):
		yield f"obj={self.obj!r}"

	def _ll_repr_pretty_(self, p):
		p.breakable()
		p.text("obj=")
		p.pretty(self.obj)

	@classmethod
	def fromul4(cls, source, node, vars):
		return cls(source, _offset(node.pos), AST.fromul4(source, node.obj, vars))

	def ul4ondump(self, encoder):
		super().ul4ondump(encoder)
		encoder.dump(self.obj)

	def ul4onload(self, decoder):
		super().ul4onload(decoder)
		self.obj = decoder.load()


@ul4on.register("de.livinglogic.vsql.not")
class Not(Unary):
	dbnodetype = "unop_not"


@ul4on.register("de.livinglogic.vsql.neg")
class Neg(Unary):
	dbnodetype = "unop_neg"


@ul4on.register("de.livinglogic.vsql.bitnot")
class BitNot(Unary):
	dbnodetype = "unop_bitnot"


@ul4on.register("de.livinglogic.vsql.if")
class If(AST):
	dbnodetype = "ternop_if"

	def __init__(self, source=None, pos=None, objif=None, objcond=None, objelse=None):
		super().__init__(source, pos)
		self.objif = objif
		self.objcond = objcond
		self.objelse = objelse

	def dbchildren(self):
		yield self.objif
		yield self.objcond
		yield self.objelse

	def _ll_repr_(self):
		yield f"objif={self.objif!r}"
		yield f"objcond={self.objcond!r}"
		yield f"objelse={self.objelse!r}"

	def _ll_repr_pretty_(self, p):
		p.breakable()
		p.text("objif=")
		p.pretty(self.objif)
		p.text("objcond=")
		p.pretty(self.objcond)
		p.text("objelse=")
		p.pretty(self.objelse)

	@classmethod
	def fromul4(cls, source, node, vars):
		return cls(
			source,
			_offset(node.pos),
			AST.fromul4(source, node.objif, vars),
			AST.fromul4(source, node.objcond, vars),
			AST.fromul4(source, node.objelse, vars),
		)

	def ul4ondump(self, encoder):
		super().ul4ondump(encoder)
		encoder.dump(self.objif)
		encoder.dump(self.objcond)
		encoder.dump(self.objelse)

	def ul4onload(self, decoder):
		super().ul4onload(decoder)
		self.objif = decoder.load()
		self.objcond = decoder.load()
		self.objelse = decoder.load()


@ul4on.register("de.livinglogic.vsql.if")
class Slice(AST):
	dbnodetype = "ternop_slice"

	def __init__(self, source=None, pos=None, obj=None, index1=None, index2=None):
		super().__init__(source, pos)
		self.obj = obj
		self.index1 = index1
		self.index2 = index2

	def dbchildren(self):
		yield self.obj
		if self.index1 is None:
			pos = self.obj.pos.stop
			yield None_(self._source, slice(pos, pos))
		else:
			pos = self.index1.stop
			yield self.index1
		if self.index2 is None:
			yield None_(self._source, slice(pos, pos))
		else:
			yield self.index2

	def _ll_repr_(self):
		yield f"obj={self.obj!r}"
		if self.index1 is not None:
			yield f"index1={self.index1!r}"
		if self.index2 is not None:
			yield f"index2={self.index2!r}"

	def _ll_repr_pretty_(self, p):
		p.breakable()
		p.text("obj=")
		p.pretty(self.obj)
		if self.index1 is not None:
			p.breakable()
			p.text("index1=")
			p.pretty(self.index1)
		if self.index2 is not None:
			p.breakable()
			p.text("index2=")
			p.pretty(self.index2)

	@classmethod
	def fromul4(cls, source, node, vars):
		return cls(
			source,
			_offset(node.pos),
			AST.fromul4(source, node.obj1, vars),
			AST.fromul4(source, node.obj2.index1, vars) if node.obj2.index1 is not None else None,
			AST.fromul4(source, node.obj2.index2, vars) if node.obj2.index2 is not None else None,
		)

	def ul4ondump(self, encoder):
		super().ul4ondump(encoder)
		encoder.dump(self.obj)
		encoder.dump(self.index1)
		encoder.dump(self.index1)

	def ul4onload(self, decoder):
		super().ul4onload(decoder)
		self.obj = decoder.load()
		self.index1 = decoder.load()
		self.index2 = decoder.load()


@ul4on.register("de.livinglogic.vsql.attr")
class Attr(AST):
	dbnodetype = "attr"

	def __init__(self, source=None, pos=None, obj=None, attrname=None):
		super().__init__(source, pos)
		self.obj = obj
		self.attrname = attrname

	@property
	def dbnodevalue(self):
		return self.attrname

	def dbchildren(self):
		yield self.obj

	def _ll_repr_(self):
		yield f"obj={self.obj!r}"
		yield f"attrname={self.attrname!r}"

	def _ll_repr_pretty_(self, p):
		p.breakable()
		p.text("obj=")
		p.pretty(self.obj)
		p.breakable()
		p.text("attrname=")
		p.pretty(self.attrname)

	@classmethod
	def fromul4(cls, source, node, vars):
		return cls(
			source,
			_offset(node.pos),
			AST.fromul4(source, node.obj1, vars),
			node.attrname,
		)

	def ul4ondump(self, encoder):
		super().ul4ondump(encoder)
		encoder.dump(self.obj)
		encoder.dump(self.attrname)

	def ul4onload(self, decoder):
		super().ul4onload(decoder)
		self.obj = decoder.load()
		self.attrname = decoder.load()


@ul4on.register("de.livinglogic.vsql.func")
class Func(AST):
	dbnodetype = "func"

	def __init__(self, source=None, pos=None, name=None, args=None):
		super().__init__(source, pos)
		self.name = name
		self.args = args

	@property
	def dbnodevalue(self):
		return self.name

	def dbchildren(self):
		yield from self.args

	def _ll_repr_(self):
		yield f"{self.name!r}"
		yield f"with {len(self.args):,} arguments"

	def _ll_repr_pretty_(self, p):
		p.text(" ")
		p.pretty(self.name)
		for (i, arg) in enumerate(self.args):
			p.breakable()
			p.text(f"args[{i}]=")
			p.pretty(arg)

	def ul4ondump(self, encoder):
		super().ul4ondump(encoder)
		encoder.dump(self.name)
		encoder.dump(self.args)

	def ul4onload(self, decoder):
		super().ul4onload(decoder)
		self.name = decoder.load()
		self.args = decoder.load()


@ul4on.register("de.livinglogic.vsql.meth")
class Meth(AST):
	dbnodetype = "meth"

	def __init__(self, source=None, pos=None, obj=None, name=None, args=None):
		super().__init__(source, pos)
		self.obj = obj
		self.name = name
		self.args = args or []

	@property
	def dbnodevalue(self):
		return self.name

	def dbchildren(self):
		yield self.obj
		yield from self.args

	def _ll_repr_(self):
		yield f"{self.name!r}"
		yield f"obj={self.obj!r}"
		yield f" with {len(self.args):,} arguments"

	def _ll_repr_pretty_(self, p):
		p.text(" ")
		p.pretty(self.name)
		p.breakable()
		p.text("obj=")
		p.pretty(self.obj)
		for (i, arg) in enumerate(self.args):
			p.breakable()
			p.text(f"args[{i}]=")
			p.pretty(arg)

	def ul4ondump(self, encoder):
		super().ul4ondump(encoder)
		encoder.dump(self.obj)
		encoder.dump(self.name)
		encoder.dump(self.args)

	def ul4onload(self, decoder):
		super().ul4onload(decoder)
		self.obj = decoder.load()
		self.name = decoder.load()
		self.args = decoder.load()


_consts = {
	bool: Bool,
	int: Int,
	float: Number,
	str: Str,
	color.Color: Color,
	datetime.date: Date,
	datetime.datetime: DateTime,
}

# Set of UL4 AST nodes that directly map to their equivalent vSQL version
_ops = {ul4c.If, ul4c.Not, ul4c.Neg, ul4c.BitNot, ul4c.List, ul4c.Set}
_ops.update(ul4c.Binary.__subclasses__())

# Create the mapping that maps the UL4 AST type to the vSQL AST type
v = vars()
_ul42vsql = {cls: v[cls.__name__] for cls in _ops}

# Remove temporary variables
del _ops, v



"""
Tests for vSQL functions.

The test are done via the Python DB interface.

To run the tests, :mod:`pytest` is required.
"""

from conftest import *


###
### Tests
###

def test_today(config_persons):
	check_vsql(config_persons, "today() >= @(2000-02-29)")

def test_now(config_persons):
	check_vsql(config_persons, "now() >= @(2000-02-29T12:34:56)")

def test_bool(config_persons):
	check_vsql(config_persons, "not bool()")

def test_bool_none(config_persons):
	check_vsql(config_persons, "not bool(None)")

def test_bool_false(config_persons):
	check_vsql(config_persons, "not bool(False)")

def test_bool_true(config_persons):
	check_vsql(config_persons, "bool(True)")

def test_bool_int_false(config_persons):
	check_vsql(config_persons, "not bool(0)")

def test_bool_int_true(config_persons):
	check_vsql(config_persons, "bool(42)")

def test_bool_number_false(config_persons):
	check_vsql(config_persons, "not bool(0.0)")

def test_bool_number_true(config_persons):
	check_vsql(config_persons, "bool(42.5)")

def test_bool_datedelta_false(config_persons):
	check_vsql(config_persons, "not bool(days(0))")

def test_bool_datedelta_true(config_persons):
	check_vsql(config_persons, "bool(days(42))")

def test_bool_datetimedelta_false(config_persons):
	check_vsql(config_persons, "not bool(minutes(0))")

def test_bool_datetimedelta_true(config_persons):
	check_vsql(config_persons, "bool(minutes(42))")

def test_bool_monthdelta_false(config_persons):
	check_vsql(config_persons, "not bool(monthdelta(0))")

def test_bool_monthdelta_true(config_persons):
	check_vsql(config_persons, "bool(monthdelta(42))")

def test_bool_date(config_persons):
	check_vsql(config_persons, "bool(@(2000-02-29))")

def test_bool_datetime(config_persons):
	check_vsql(config_persons, "bool(@(2000-02-29T12:34:56))")

def test_bool_color(config_persons):
	check_vsql(config_persons, "bool(#fff)")

def test_bool_str_false(config_persons):
	check_vsql(config_persons, "not bool('')")

def test_bool_str_true(config_persons):
	check_vsql(config_persons, "bool('gurk')")

def test_bool_intlist(config_persons):
	check_vsql(config_persons, "bool([42])")

def test_bool_numberlist(config_persons):
	check_vsql(config_persons, "bool([42.5])")

def test_bool_strlist(config_persons):
	check_vsql(config_persons, "bool(['gurk'])")

def test_bool_datelist(config_persons):
	check_vsql(config_persons, "bool([today()])")

def test_bool_datetimelist(config_persons):
	check_vsql(config_persons, "bool([now()])")

def test_bool_intset(config_persons):
	check_vsql(config_persons, "bool({42})")

def test_bool_numberset(config_persons):
	check_vsql(config_persons, "bool({42.5})")

def test_bool_strset(config_persons):
	check_vsql(config_persons, "bool({'gurk'})")

def test_bool_dateset(config_persons):
	check_vsql(config_persons, "bool({today()})")

def test_bool_datetimeset(config_persons):
	check_vsql(config_persons, "bool({now()})")

def test_int(config_persons):
	check_vsql(config_persons, "not int()")

def test_int_bool_false(config_persons):
	check_vsql(config_persons, "not int(False)")

def test_int_bool_true(config_persons):
	check_vsql(config_persons, "int(True)")

def test_int_int(config_persons):
	check_vsql(config_persons, "int(42) == 42")

def test_int_number(config_persons):
	check_vsql(config_persons, "int(42.4) == 42")

def test_int_str_ok(config_persons):
	check_vsql(config_persons, "int('42') == 42")

def test_int_str_bad(config_persons):
	check_vsql(config_persons, "int('42.5') is None")

def test_float(config_persons):
	check_vsql(config_persons, "float() == 0.0")

def test_float_bool_false(config_persons):
	check_vsql(config_persons, "float(False) == 0.0")

def test_float_bool_true(config_persons):
	check_vsql(config_persons, "float(True) == 1.0")

def test_float_int(config_persons):
	check_vsql(config_persons, "float(42) == 42.0")

def test_float_number(config_persons):
	check_vsql(config_persons, "float(42.5) == 42.5")

def test_str(config_persons):
	check_vsql(config_persons, "str() is None")

def test_str_bool_false(config_persons):
	check_vsql(config_persons, "str(False) == 'False'")

def test_str_bool_true(config_persons):
	check_vsql(config_persons, "str(True) == 'True'")

def test_str_int(config_persons):
	check_vsql(config_persons, "str(-42) == '-42'")

def test_str_number(config_persons):
	check_vsql(config_persons, "str(42.0) == '42.0' and str(-42.5) == '-42.5'")

def test_str_str(config_persons):
	check_vsql(config_persons, "str('foo') == 'foo'")

def test_str_date(config_persons):
	check_vsql(config_persons, "str(@(2000-02-29)) == '2000-02-29'")

def test_str_datetime(config_persons):
	check_vsql(config_persons, "str(@(2000-02-29T12:34:56)) == '2000-02-29 12:34:56'")

def test_str_datedelta_1(config_persons):
	check_vsql(config_persons, "str(days(1)) == '1 day'")

def test_str_datedelta_2(config_persons):
	check_vsql(config_persons, "str(days(42)) == '42 days'")

def test_str_datetimedelta_1(config_persons):
	check_vsql(config_persons, "str(seconds(42)) == '0:00:42'")

def test_str_datetimedelta_2(config_persons):
	check_vsql(config_persons, "str(minutes(42)) == '0:42:00'")

def test_str_datetimedelta_3(config_persons):
	check_vsql(config_persons, "str(hours(17) + minutes(23)) == '17:23:00'")

def test_str_datetimedelta_4(config_persons):
	check_vsql(config_persons, "str(hours(42) + seconds(0)) == '1 day, 18:00:00'")

def test_str_datetimedelta_5(config_persons):
	check_vsql(config_persons, "str(days(42) + seconds(0)) == '42 days, 0:00:00'")

def test_str_datetimedelta_6(config_persons):
	check_vsql(config_persons, "str(days(42) + hours(17) + minutes(23)) == '42 days, 17:23:00'")

def test_str_monthdelta_1(config_persons):
	check_vsql(config_persons, "str(monthdelta(0)) == '0 months'")

def test_str_monthdelta_2(config_persons):
	check_vsql(config_persons, "str(monthdelta(1)) == '1 month'")

def test_str_monthdelta_3(config_persons):
	check_vsql(config_persons, "str(monthdelta(42)) == '42 months'")

def test_str_color_1(config_persons):
	check_vsql(config_persons, "str(#000f) == '#000'")

def test_str_color_2(config_persons):
	check_vsql(config_persons, "str(#fff0) == 'rgba(255, 255, 255, 0.000)'")

def test_str_color_3(config_persons):
	check_vsql(config_persons, "str(#123456) == '#123456'")

def test_str_color_4(config_persons):
	check_vsql(config_persons, "str(#12345678) == 'rgba(18, 52, 86, 0.471)'")

def test_str_intlist(config_persons):
	check_vsql(config_persons, "str([1, 2, 3, None]) == '[1, 2, 3, None]'")

def test_str_numberlist(config_persons):
	check_vsql(config_persons, "str([1.2, 3.4, 5.6, None]) == '[1.2, 3.4, 5.6, None]'")

def test_str_strlist(config_persons):
	check_vsql(config_persons, "str(['foo', 'bar', None]) == '[\\'foo\\', \\'bar\\', None]'")

def test_str_datelist(config_persons):
	check_vsql(config_persons, "str([@(2000-02-29), None]) == '[@(2000-02-29), None]'")

def test_str_datetimelist(config_persons):
	check_vsql(config_persons, "str([@(2000-02-29T12:34:56), None]) == '[@(2000-02-29T12:34:56), None]'")

# For the set test only include one non-``None`` value,
# as the order of the other elements is undefined

def test_str_intset(config_persons):
	check_vsql(config_persons, "str({1, None}) == '{1, None}'")

def test_str_numberset(config_persons):
	check_vsql(config_persons, "str({1.2, None}) == '{1.2, None}'")

def test_str_strset(config_persons):
	check_vsql(config_persons, "str({'foo', None}) == '{\\'foo\\', None}'")

def test_str_dateset(config_persons):
	check_vsql(config_persons, "str({@(2000-02-29), None}) == '{@(2000-02-29), None}'")

def test_str_datetimeset(config_persons):
	check_vsql(config_persons, "str({@(2000-02-29T12:34:56), None}) == '{@(2000-02-29T12:34:56), None}'")

def test_repr_none(config_persons):
	check_vsql(config_persons, "repr(None) == 'None'")

def test_repr_bool_false(config_persons):
	check_vsql(config_persons, "repr(False) == 'False'")

def test_repr_bool_True(config_persons):
	check_vsql(config_persons, "repr(True) == 'True'")

def test_repr_int(config_persons):
	check_vsql(config_persons, "repr(-42) == '-42'")

def test_repr_number_1(config_persons):
	check_vsql(config_persons, "repr(42.0) == '42.0'")

def test_repr_number_2(config_persons):
	check_vsql(config_persons, "repr(-42.5) == '-42.5'")

def test_repr_str(config_persons):
	check_vsql(config_persons, "repr('foo\"bar') == '\\'foo\\\"bar\\''")

def test_repr_date(config_persons):
	check_vsql(config_persons, "repr(@(2000-02-29)) == '@(2000-02-29)'")

def test_repr_datetime(config_persons):
	check_vsql(config_persons, "repr(@(2000-02-29T12:34:56)) == '@(2000-02-29T12:34:56)'")

def test_repr_datedelta_1(config_persons):
	check_vsql(config_persons, "repr(days(1)) == 'timedelta(1)'")

def test_repr_datedelta_2(config_persons):
	check_vsql(config_persons, "repr(days(42)) == 'timedelta(42)'")

def test_repr_datetimedelta_1(config_persons):
	# FIXME: Oracle doesn't have enough precision for seconds
	check_vsql(config_persons, "repr(seconds(42)) == 'timedelta(0, 42)'")

def test_repr_datetimedelta_2(config_persons):
	check_vsql(config_persons, "repr(minutes(42)) == 'timedelta(0, 2520)'")

def test_repr_datetimedelta_3(config_persons):
	check_vsql(config_persons, "repr(hours(17) + minutes(23)) == 'timedelta(0, 62580)'")

def test_repr_datetimedelta_4(config_persons):
	check_vsql(config_persons, "repr(hours(42) + seconds(0)) == 'timedelta(1, 64800)'")

def test_repr_datetimedelta_5(config_persons):
	check_vsql(config_persons, "repr(days(42) + seconds(0)) == 'timedelta(42)'")

def test_repr_datetimedelta_6(config_persons):
	check_vsql(config_persons, "repr(days(42) + hours(17) + minutes(23)) == 'timedelta(42, 62580)'")

def test_repr_monthdelta(config_persons):
	check_vsql(config_persons, "repr(monthdelta(42)) == 'monthdelta(42)'")

def test_repr_color_1(config_persons):
	check_vsql(config_persons, "repr(#000) == '#000'")

def test_repr_color_2(config_persons):
	check_vsql(config_persons, "repr(#369c) == '#369c'")

def test_repr_color_3(config_persons):
	check_vsql(config_persons, "repr(#123456) == '#123456'")

def test_repr_color_4(config_persons):
	check_vsql(config_persons, "repr(#12345678) == '#12345678'")

def test_repr_intlist(config_persons):
	check_vsql(config_persons, "repr([1, 2, 3, None]) == '[1, 2, 3, None]'")

def test_repr_numberlist(config_persons):
	check_vsql(config_persons, "repr([1.2, 3.4, 5.6, None]) == '[1.2, 3.4, 5.6, None]'")

def test_repr_strlist(config_persons):
	check_vsql(config_persons, "repr(['foo', 'bar', None]) == '[\\'foo\\', \\'bar\\', None]'")

def test_repr_datelist(config_persons):
	check_vsql(config_persons, "repr([@(2000-02-29), None]) == '[@(2000-02-29), None]'")

def test_repr_datetimelist(config_persons):
	check_vsql(config_persons, "repr([@(2000-02-29T12:34:56), None]) == '[@(2000-02-29T12:34:56), None]'")

# For the set test only include one non-``None`` value,
# as the order of the other elements is undefined

def test_repr_intset(config_persons):
	check_vsql(config_persons, "repr({1, None}) == '{1, None}'")

def test_repr_numberset(config_persons):
	check_vsql(config_persons, "repr({1.2, None}) == '{1.2, None}'")

def test_repr_strset(config_persons):
	check_vsql(config_persons, "repr({'foo', None}) == '{\\\'foo\\\', None}'")

def test_repr_dateset(config_persons):
	check_vsql(config_persons, "repr({@(2000-02-29), None}) == '{@(2000-02-29), None}'")

def test_repr_datetimeset(config_persons):
	check_vsql(config_persons, "repr({@(2000-02-29T12:34:56), None}) == '{@(2000-02-29T12:34:56), None}'")

def test_date_int(config_persons):
	check_vsql(config_persons, "date(2000, 2, 29) == @(2000-02-29)")

def test_date_datetime(config_persons):
	check_vsql(config_persons, "date(@(2000-02-29T12:34:56)) == @(2000-02-29)")

def test_datetime_int3(config_persons):
	check_vsql(config_persons, "datetime(2000, 2, 29) == @(2000-02-29T)")

def test_datetime_int4(config_persons):
	check_vsql(config_persons, "datetime(2000, 2, 29, 12) == @(2000-02-29T12:00:00)")

def test_datetime_int5(config_persons):
	check_vsql(config_persons, "datetime(2000, 2, 29, 12, 34) == @(2000-02-29T12:34:00)")

def test_datetime_int6(config_persons):
	check_vsql(config_persons, "datetime(2000, 2, 29, 12, 34, 56) == @(2000-02-29T12:34:56)")

def test_len_str1(config_persons):
	check_vsql(config_persons, "len('') == 0")

def test_len_str2(config_persons):
	check_vsql(config_persons, "len('gurk') == 4")

def test_len_str3(config_persons):
	check_vsql(config_persons, "len('\\t\\n') == 2")

def test_len_intlist(config_persons):
	check_vsql(config_persons, "len([1, 2, 3]) == 3")

def test_len_numberlist(config_persons):
	check_vsql(config_persons, "len([1.2, 3.4, 5.6]) == 3")

def test_len_strlist(config_persons):
	check_vsql(config_persons, "len(['foo', 'bar', 'baz']) == 3")

def test_len_datelist(config_persons):
	check_vsql(config_persons, "len([@(2000-02-29)]) == 1")

def test_len_datetimelist(config_persons):
	check_vsql(config_persons, "len([@(2000-02-29T12:34:56)]) == 1")

def test_len_intset(config_persons):
	check_vsql(config_persons, "len({1, 1, 2, 2, 3, 3, None, None}) == 4")

def test_len_numberset(config_persons):
	check_vsql(config_persons, "len({1.2, 3.4, 5.6, None, 1.2, 3.4, 5.6, None}) == 4")

def test_len_strset(config_persons):
	check_vsql(config_persons, "len({'foo', 'bar', 'baz', None, 'foo', 'bar', 'baz'}) == 4")

def test_len_dateset(config_persons):
	check_vsql(config_persons, "len({@(2000-02-29), @(2000-02-29), @(2000-03-21), None}) == 3")

def test_len_datetimeset(config_persons):
	check_vsql(config_persons, "len({@(2000-02-29T12:34:56), None, @(2000-02-29T12:34:56), None, @(2000-02-29T11:22:33)}) == 3")

def test_timedelta(config_persons):
	check_vsql(config_persons, "not timedelta()")

def test_timedelta_int1(config_persons):
	check_vsql(config_persons, "timedelta(42)")

def test_timedelta_int2(config_persons):
	check_vsql(config_persons, "timedelta(42, 12)")

def test_monthdelta(config_persons):
	check_vsql(config_persons, "not monthdelta()")

def test_monthdelta_int(config_persons):
	check_vsql(config_persons, "monthdelta(42)")

def test_years(config_persons):
	check_vsql(config_persons, "years(25)")

def test_months(config_persons):
	check_vsql(config_persons, "months(3)")

def test_weeks(config_persons):
	check_vsql(config_persons, "weeks(3)")

def test_days(config_persons):
	check_vsql(config_persons, "days(12)")

def test_hours(config_persons):
	check_vsql(config_persons, "hours(8)")

def test_minutes(config_persons):
	check_vsql(config_persons, "minutes(45)")

def test_seconds(config_persons):
	check_vsql(config_persons, "seconds(60)")

def test_md5(config_persons):
	check_vsql(config_persons, "md5('gurk') == '4b5b6a3fa4af2541daa569277c7ff4c5'")

def test_random(config_persons):
	check_vsql(config_persons, "random() + 1")

def test_randrange(config_persons):
	check_vsql(config_persons, "randrange(1, 10)")

def test_seq(config_persons):
	check_vsql(config_persons, "seq()")

def test_rgb1(config_persons):
	check_vsql(config_persons, "rgb(0.2, 0.4, 0.6) == #369")

def test_rgb2(config_persons):
	check_vsql(config_persons, "rgb(0.2, 0.4, 0.6, 0.8) == #369c")

def test_list_str(config_persons):
	check_vsql(config_persons, "list('gurk') == ['g', 'u', 'r', 'k']")

def test_list_intlist(config_persons):
	check_vsql(config_persons, "list([1, 2, 3]) == [1, 2, 3]")

def test_list_numberlist(config_persons):
	check_vsql(config_persons, "list([1.2, 3.4, 5.6]) == [1.2, 3.4, 5.6]")

def test_list_strlist(config_persons):
	check_vsql(config_persons, "list(['foo', 'bar', 'baz', None]) == ['foo', 'bar', 'baz', None]")

def test_list_datelist(config_persons):
	check_vsql(config_persons, "list([@(2000-02-29), @(2000-03-01), None]) == [@(2000-02-29), @(2000-03-01), None]")

def test_list_datetimelist(config_persons):
	check_vsql(config_persons, "list([@(2000-02-29T12:34:56), @(2000-02-29T11:22:33), None]) == [@(2000-02-29T12:34:56), @(2000-02-29T11:22:33), None]")

def test_list_intset(config_persons):
	check_vsql(config_persons, "list({1, None}) == [1, None]")

def test_list_numberset(config_persons):
	check_vsql(config_persons, "list({1.2, None}) == [1.2, None]")

def test_list_strset(config_persons):
	check_vsql(config_persons, "list({'foo', None}) == ['foo', None]")

def test_list_dateset(config_persons):
	check_vsql(config_persons, "list({@(2000-02-29), None}) == [@(2000-02-29), None]")

def test_list_datetimeset(config_persons):
	check_vsql(config_persons, "list({@(2000-02-29T12:34:56), None}) == [@(2000-02-29T12:34:56), None]")

def test_set_str(config_persons):
	check_vsql(config_persons, "set('mississippi') == {'i', 'm', 'p', 's'}")

def test_set_intlist(config_persons):
	check_vsql(config_persons, "set([1, 2, 3, 2, 1, None]) == {1, 2, 3, None}")

def test_set_numberlist(config_persons):
	check_vsql(config_persons, "set([1.2, 3.4, 5.6, 3.4, 1.2, None]) == {1.2, 3.4, 5.6, None}")

def test_set_strlist(config_persons):
	check_vsql(config_persons, "set(['foo', 'bar', 'baz', None, 'baz', 'bar', 'foo']) == {'foo', 'bar', 'baz', None}")

def test_set_datelist(config_persons):
	check_vsql(config_persons, "set([@(2000-02-29), @(2000-03-01), None, @(2000-03-01), @(2000-02-29)]) == {@(2000-02-29), @(2000-03-01), None}")

def test_set_datetimelist(config_persons):
	check_vsql(config_persons, "set([@(2000-02-29T12:34:56), @(2000-02-29T11:22:33), @(2000-02-29T11:22:33), None, @(2000-02-29T12:34:56)]) == {@(2000-02-29T12:34:56), @(2000-02-29T11:22:33), None}")

def test_set_intset(config_persons):
	check_vsql(config_persons, "set({1, None}) == {1, None}")

def test_set_numberset(config_persons):
	check_vsql(config_persons, "set({1.2, None}) == {1.2, None}")

def test_set_strset(config_persons):
	check_vsql(config_persons, "set({'foo', None}) == {'foo', None}")

def test_set_dateset(config_persons):
	check_vsql(config_persons, "set({@(2000-02-29), None}) == {@(2000-02-29), None}")

def test_set_datetimeset(config_persons):
	check_vsql(config_persons, "set({@(2000-02-29T12:34:56), None}) == {@(2000-02-29T12:34:56), None}")



"""
Tests for vSQL.

The test are done via the Python DB interface.

To run the tests, :mod:`pytest` is required.
"""

import operator

from conftest import *


template_unsorted_persons = """
	<?whitespace strip?>
	<?for (f, r) in isfirst(datasources.persons.app.records.values())?>
		<?if not f?>;<?end if?>
		<?print r.v_firstname?> <?print r.v_lastname?>
	<?end for?>
"""

template_sorted_persons = """
	<?whitespace strip?>
	<?def key(r)?>
		<?return r.v_lastname?>
	<?end def?>
	<?for (f, r) in isfirst(sorted(datasources.persons.app.records.values(), key))?>
		<?if not f?>;<?end if?>
		<?print r.v_firstname?> <?print r.v_lastname?>
	<?end for?>
"""

template_unsorted_children = """
	<?whitespace strip?>
	<?for (f, r) in isfirst(datasources.fieldsofactivity.app.records[id].c_children.values())?>
		<?if not f?>;<?end if?>
		<?print r.v_name?>
	<?end for?>
"""

template_sorted_children = """
	<?whitespace strip?>
	<?def key(r)?>
		<?return r.v_name?>
	<?end def?>
	<?for (f, r) in isfirst(sorted(datasources.fieldsofactivity.app.records[id].c_children.values(), key))?>
		<?if not f?>;<?end if?>
		<?print r.v_name?>
	<?end for?>
"""


###
### Tests
###

def test_global_variables(config_persons):
	c = config_persons

	handler = PythonDB()

	source = f"""
		<?whitespace strip?>
		<?for r in datasources.fieldsofactivity.app.records.values()?>
			;<?print r.v_name?>
			<?for r2 in r.c_children.values()?>
				;<?print r2.v_name?>
			<?end for?>
		<?end for?>
	"""

	vt = handler.make_viewtemplate(
		la.DataSource(
			la.DataOrder(expression="r.v_name"),
			la.DataOrder(expression="app.p_str_value.value"),
			la.DataOrder(expression="user.email"),
			la.DataOrder(expression="record.id"),
			la.DataOrder(expression="params.str.nix"),
			la.DataOrder(expression="len(params.strlist.nix)"),
			la.DataSourceChildren(
				la.DataOrder(expression="r.v_name"),
				la.DataOrder(expression="app.p_str_value.value"),
				la.DataOrder(expression="user.email"),
				la.DataOrder(expression="record.id"),
				la.DataOrder(expression="params.str.nix"),
				la.DataOrder(expression="len(params.strlist.nix)"),
				identifier="children",
				control=c.apps.fields.c_parent,
				filter="r.app.id != app.p_app_value.id and user.email is not None and record.id is None and params.str.nix is None and len(params.strlist.nix) == 0",
			),
			identifier="fieldsofactivity",
			app=c.apps.fields,
			recordfilter="r.app.id != app.p_app_value.id and user.email is not None and record.id is None and params.str.nix is None and len(params.strlist.nix) == 0",
			includeparams=True,
		),
		identifier="vsql_global_variables",
		source=source,
	)

	output = handler.renders(person_app_id, template=vt.identifier)

	expected = []
	key = operator.attrgetter("v_name")
	for a in sorted(c.areas.values(), key=key):
		expected.append(a.v_name)
		for a2 in sorted((a2 for a2 in c.areas.values() if a2.v_parent is a), key=key):
			expected.append(a2.v_name)
	expected = ";" + ";".join(expected)

	assert expected == output


def test_datasource_appfilter(config_persons):
	handler = PythonDB()

	source = """
		<?whitespace strip?>
		<?print repr(datasources.all.app)?>
		;
		<?for a in datasources.all.apps.values()?>
			<?print a.id?>
		<?end for?>
	"""

	vt = handler.make_viewtemplate(
		la.DataSource(
			identifier="all",
			appfilter=f"a.uuid == '{person_app_id}'"
		),
		identifier="vsql_datasource_appfilter",
		source=source,
	)

	output = handler.renders(person_app_id, template=vt.identifier)

	assert f"None;{person_app_id}" == output


def test_datasource_recordfilter(config_persons):
	c = config_persons

	handler = PythonDB()

	vt = handler.make_viewtemplate(
		la.DataSource(
			identifier="persons",
			app=c.apps.persons,
			recordfilter="r.v_lastname == 'Einstein'",
		),
		identifier="vsql_datasource_recordfilter",
		source=template_unsorted_persons,
	)
	output = handler.renders(person_app_id, template=vt.identifier)

	assert "Albert Einstein" == output


def test_datasource_recordfilter_param_str(config_persons):
	c = config_persons

	handler = PythonDB()

	vt = handler.make_viewtemplate(
		la.DataSource(
			identifier="persons",
			app=c.apps.persons,
			recordfilter="r.v_lastname == params.str.lastname",
		),
		identifier="vsql_datasource_recordfilter_param_str",
		source=template_unsorted_persons,
	)
	output = handler.renders(person_app_id, template=vt.identifier, lastname="Curie")

	assert "Marie Curie" == output


def test_datasource_recordfilter_param_int(config_persons):
	c = config_persons

	handler = PythonDB()

	vt = handler.make_viewtemplate(
		la.DataSource(
			identifier="persons",
			app=c.apps.persons,
			recordfilter="r.v_date_of_birth.year == params.int.year",
		),
		identifier="vsql_datasource_recordfilter_param_int",
		source=template_unsorted_persons,
	)
	output = handler.renders(person_app_id, template=vt.identifier, year="1935")

	assert "Elvis Presley" == output


def test_datasource_recordfilter_param_date(config_persons):
	c = config_persons

	handler = PythonDB()

	vt = handler.make_viewtemplate(
		la.DataSource(
			identifier="persons",
			app=c.apps.persons,
			recordfilter="r.v_date_of_birth == params.date.date_of_birth",
		),
		identifier="vsql_datasource_recordfilter_param_date",
		source=template_unsorted_persons,
	)
	output = handler.renders(person_app_id, template=vt.identifier, date_of_birth="1926-06-01")

	assert "Marilyn Monroe" == output


def test_datasource_recordfilter_param_datetime(config_persons):
	c = config_persons

	handler = PythonDB()

	vt = handler.make_viewtemplate(
		la.DataSource(
			identifier="persons",
			app=c.apps.persons,
			recordfilter="datetime(r.v_date_of_birth) + hours(12) + minutes(34) + seconds(56) == params.datetime.date_of_birth",
		),
		identifier="vsql_datasource_recordfilter_param_datetime",
		source=template_unsorted_persons,
	)
	output = handler.renders(person_app_id, template=vt.identifier, date_of_birth="1926-06-01T12:34:56")

	assert "Marilyn Monroe" == output


def test_datasource_recordfilter_param_strlist(config_persons):
	c = config_persons

	handler = PythonDB()

	vt = handler.make_viewtemplate(
		la.DataSource(
			identifier="persons",
			app=c.apps.persons,
			recordfilter="r.v_lastname in params.strlist.lastname",
		),
		identifier="vsql_datasource_recordfilter_param_strlist",
		source=template_sorted_persons,
	)
	output = handler.renders(person_app_id, template=vt.identifier, lastname=["Gauß", "Riemann"])

	assert "Carl Friedrich Gauß;Bernhard Riemann" == output


def test_datasource_recordfilter_param_intlist(config_persons):
	c = config_persons

	handler = PythonDB()

	vt = handler.make_viewtemplate(
		la.DataSource(
			identifier="persons",
			app=c.apps.persons,
			recordfilter="r.v_date_of_birth.year in params.intlist.year",
		),
		identifier="vsql_datasource_recordfilter_param_intlist",
		source=template_sorted_persons,
	)
	output = handler.renders(person_app_id, template=vt.identifier, year=["1826", "1777"])

	assert "Carl Friedrich Gauß;Bernhard Riemann" == output


def test_datasource_recordfilter_param_datelist(config_persons):
	c = config_persons

	handler = PythonDB()

	vt = handler.make_viewtemplate(
		la.DataSource(
			identifier="persons",
			app=c.apps.persons,
			recordfilter="r.v_date_of_birth in params.datelist.date_of_birth",
		),
		identifier="vsql_datasource_recordfilter_param_datelist",
		source=template_sorted_persons,
	)
	output = handler.renders(person_app_id, template=vt.identifier, date_of_birth=["1826-06-17", "1777-04-30"])

	assert "Carl Friedrich Gauß;Bernhard Riemann" == output


def test_datasource_recordfilter_param_datetimelist(config_persons):
	c = config_persons

	handler = PythonDB()

	vt = handler.make_viewtemplate(
		la.DataSource(
			identifier="persons",
			app=c.apps.persons,
			recordfilter="datetime(r.v_date_of_birth) + hours(12) + minutes(34) + seconds(56) == params.datetimelist.date_of_birth[0]",
		),
		identifier="vsql_datasource_recordfilter_param_datelist",
		source=template_sorted_persons,
	)
	output = handler.renders(person_app_id, template=vt.identifier, date_of_birth=["1777-04-30T12:34:56"])

	assert "Carl Friedrich Gauß" == output


def test_datasource_recordfilter_appparam_int(config_persons):
	c = config_persons

	handler = PythonDB()

	vt = handler.make_viewtemplate(
		la.DataSource(
			identifier="persons",
			app=c.apps.persons,
			recordfilter="r.v_date_of_birth.year == app.p_int_value.value",
		),
		identifier="vsql_datasource_recordfilter_appparam_int",
		source=template_sorted_persons,
	)
	output = handler.renders(person_app_id, template=vt.identifier)

	assert "Carl Friedrich Gauß" == output


def test_datasource_sort_asc_nullsfirst(config_persons):
	c = config_persons

	handler = PythonDB()

	vt = handler.make_viewtemplate(
		la.DataSource(
			la.DataOrder(expression="r.v_date_of_death", direction="asc", nulls="first"),
			identifier="persons",
			app=c.apps.persons,
			recordfilter="r.v_lastname in ['Knuth', 'Gauß', 'Einstein', 'Riemann']",
		),
		identifier="vsql_datasource_sort_asc_nullsfirst",
		source=template_unsorted_persons,
	)
	output = handler.renders(person_app_id, template=vt.identifier)

	assert "Donald Knuth;Carl Friedrich Gauß;Bernhard Riemann;Albert Einstein" == output


def test_datasource_sort_asc_nullslast(config_persons):
	c = config_persons

	handler = PythonDB()

	vt = handler.make_viewtemplate(
		la.DataSource(
			la.DataOrder(expression="r.v_date_of_death", direction="asc", nulls="last"),
			identifier="persons",
			app=c.apps.persons,
			recordfilter="r.v_lastname in ['Knuth', 'Gauß', 'Einstein', 'Riemann']",
		),
		identifier="vsql_datasource_sort_asc_nullslast",
		source=template_unsorted_persons,
	)

	output = handler.renders(person_app_id, template=vt.identifier)

	assert "Carl Friedrich Gauß;Bernhard Riemann;Albert Einstein;Donald Knuth" == output


def test_datasource_sort_desc_nullsfirst(config_persons):
	c = config_persons

	handler = PythonDB()

	vt = handler.make_viewtemplate(
		la.DataSource(
			la.DataOrder(expression="r.v_date_of_death", direction="desc", nulls="first"),
			identifier="persons",
			app=c.apps.persons,
			recordfilter="r.v_lastname in ['Knuth', 'Gauß', 'Einstein', 'Riemann']",
		),
		identifier="vsql_datasource_sort_desc_nullsfirst",
		source=template_unsorted_persons,
	)

	output = handler.renders(person_app_id, template=vt.identifier)

	assert "Donald Knuth;Albert Einstein;Bernhard Riemann;Carl Friedrich Gauß" == output


def test_datasource_sort_desc_nullslast(config_persons):
	c = config_persons

	handler = PythonDB()

	vt = handler.make_viewtemplate(
		la.DataSource(
			la.DataOrder(expression="r.v_date_of_death", direction="desc", nulls="last"),
			identifier="persons",
			app=c.apps.persons,
			recordfilter="r.v_lastname in ['Knuth', 'Gauß', 'Einstein', 'Riemann']",
		),
		identifier="vsql_datasource_sort_desc_nullslast",
		source=template_unsorted_persons,
	)

	output = handler.renders(person_app_id, template=vt.identifier)

	assert "Albert Einstein;Bernhard Riemann;Carl Friedrich Gauß;Donald Knuth" == output


def test_datasource_masterdetail_recordfilter(config_persons):
	c = config_persons

	handler = PythonDB()

	source = f"""
		<?whitespace strip?>
		<?print all(r.v_parent is None for r in datasources.fieldsofactivity.app.records.values())?>
		<?for id in ['{c.areas.science.id}', '{c.areas.art.id}']?>
			;{template_sorted_children}
		<?end for?>
	"""

	vt = handler.make_viewtemplate(
		la.DataSource(
			la.DataSourceChildren(
				identifier="children",
				control=c.apps.fields.c_parent,
				filter="len(r.v_name) >= 6",
			),
			identifier="fieldsofactivity",
			app=c.apps.fields,
			recordfilter="r.v_parent.id is None",
		),
		identifier="vsql_datasource_masterdetail_recordfilter",
		source=source,
	)

	output = handler.renders(person_app_id, template=vt.identifier)

	assert "True;Computer science;Mathematics;Physics;Literature" == output


def test_datasource_masterdetail_sort_asc(config_persons):
	c = config_persons

	handler = PythonDB()

	source = f"""
		<?whitespace strip?>
		<?print all(r.v_parent is None for r in datasources.fieldsofactivity.app.records.values())?>
		<?for id in ['{c.areas.science.id}', '{c.areas.art.id}']?>
			;{template_unsorted_children}
		<?end for?>
	"""

	vt = handler.make_viewtemplate(
		la.DataSource(
			la.DataSourceChildren(
				la.DataOrder(expression="r.v_name", direction="asc", nulls="first"),
				identifier="children",
				control=c.apps.fields.c_parent,
			),
			identifier="fieldsofactivity",
			app=c.apps.fields,
			recordfilter="r.v_parent.id is None",
		),
		identifier="vsql_datasource_masterdetail_sort_asc",
		source=source,
	)

	output = handler.renders(person_app_id, template=vt.identifier)

	assert "True;Computer science;Mathematics;Physics;Film;Literature;Music" == output


def test_datasource_masterdetail_sort_desc(config_persons):
	c = config_persons

	handler = PythonDB()

	source = f"""
		<?whitespace strip?>
		<?print all(r.v_parent is None for r in datasources.fieldsofactivity.app.records.values())?>
		<?for id in ['{c.areas.science.id}', '{c.areas.art.id}']?>
			;{template_unsorted_children}
		<?end for?>
	"""

	vt = handler.make_viewtemplate(
		la.DataSource(
			la.DataSourceChildren(
				la.DataOrder(expression="r.v_name", direction="desc", nulls="first"),
				identifier="children",
				control=c.apps.fields.c_parent,
			),
			identifier="fieldsofactivity",
			app=c.apps.fields,
			recordfilter="r.v_parent.id is None",
		),
		identifier="vsql_datasource_masterdetail_sort_desc",
		source=source,
	)

	output = handler.renders(person_app_id, template=vt.identifier)

	assert "True;Physics;Mathematics;Computer science;Music;Literature;Film" == output


def test_color_attributes(config_persons):
	c = config_persons

	handler = PythonDB()

	source = f"""
		<?whitespace strip?>
		<?print len(datasources.persons.app.records)?>
	"""

	vt = handler.make_viewtemplate(
		la.DataSource(
			identifier="persons",
			app=c.apps.persons,
			recordfilter="app.p_color_value.value.r == 0x33 and app.p_color_value.value.g == 0x66 and app.p_color_value.value.b == 0x99 and app.p_color_value.value.a == 0xcc",
		),
		identifier="vsql_color_attributes",
		source=source,
	)

	output = handler.renders(person_app_id, template=vt.identifier)

	assert "0" != output


def test_color_methods(config_persons):
	c = config_persons

	handler = PythonDB()

	source = f"""
		<?whitespace strip?>
		<?print len(datasources.persons.app.records)?>
	"""

	vt = handler.make_viewtemplate(
		la.DataSource(
			identifier="persons",
			app=c.apps.persons,
			recordfilter="app.p_color_value.value.lum() == 0.4",
		),
		identifier="vsql_color_methods",
		source=source,
	)

	output = handler.renders(person_app_id, template=vt.identifier)

	assert "0" != output


def test_repr_color(config_persons):
	c = config_persons

	handler = PythonDB()

	source = f"""
		<?whitespace strip?>
		<?print len(datasources.persons.app.records)?>
	"""

	vt = handler.make_viewtemplate(
		la.DataSource(
			identifier="persons",
			app=c.apps.persons,
			recordfilter='repr(#369) == "#369" and repr(#369c) == "#369c" and repr(#123456) == "#123456" and repr(#12345678) == "#12345678"',
		),
		identifier="vsql_repr_color",
		source=source,
	)

	output = handler.renders(person_app_id, template=vt.identifier)

	assert "0" != output


def test_datasource_paging(config_persons):
	c = config_persons

	handler = PythonDB()

	vt = handler.make_viewtemplate(
		la.DataSource(
			la.DataOrder("r.v_lastname"),
			la.DataOrder("r.v_firstname"),
			identifier="persons",
			app=c.apps.persons,
		),
		identifier="vsql_datasource_paging",
		source=template_unsorted_persons,
	)

	output = handler.renders(
		person_app_id,
		template=vt.identifier,
		**{"la-ds-persons-paging": "0_2"},
	)

	assert "Muhammad Ali;Marie Curie" == output


def test_datasourcechildren_paging(config_persons):
	c = config_persons

	handler = PythonDB()

	source = f"""
		<?whitespace strip?>
		<?for (f, r) in isfirst(datasources.fieldsofactivity.app.records['{c.areas.film.id}'].c_persons.values())?>
			<?if not f?>;<?end if?><?print r.v_firstname?> <?print r.v_lastname?>
		<?end for?>
	"""

	vt = handler.make_viewtemplate(
		la.DataSource(
			la.DataOrder("r.v_name"),
			la.DataSourceChildren(
				la.DataOrder("r.v_lastname"),
				la.DataOrder("r.v_firstname"),
				identifier="persons",
				control=c.apps.persons.c_field_of_activity,
			),
			identifier="fieldsofactivity",
			app=c.apps.fields,
		),
		identifier="vsql_datasourcechildren_paging",
		source=source,
	)

	output = handler.renders(
		person_app_id,
		template=vt.identifier,
		**{f"la-dsc-fieldsofactivity-{c.areas.film.id}-persons-paging": "1_1"},
	)

	assert "Ronald Reagan" == output



"""
Tests for the vSQL equality comparision operator ``==``.

The test are done via the Python DB interface.

To run the tests, :mod:`pytest` is required.
"""

from conftest import *


###
### Tests
###

def test_bool_none_false(config_persons):
	check_vsql(config_persons, "app.p_bool_none.value == None")

def test_bool_none_true(config_persons):
	check_vsql(config_persons, "not (app.p_bool_true.value == None)")

def test_int_none_false(config_persons):
	check_vsql(config_persons, "app.p_int_none.value == None")

def test_int_none_true(config_persons):
	check_vsql(config_persons, "not (app.p_int_value.value == None)")

def test_number_none_false(config_persons):
	check_vsql(config_persons, "app.p_number_none.value == None")

def test_number_none_true(config_persons):
	check_vsql(config_persons, "not (app.p_number_value.value == None)")

def test_str_none_false(config_persons):
	check_vsql(config_persons, "app.p_str_none.value == None")

def test_str_none_true(config_persons):
	check_vsql(config_persons, "not (app.p_str_value.value == None)")

def test_date_none_false(config_persons):
	check_vsql(config_persons, "app.p_date_none.value == None")

def test_date_none_true(config_persons):
	check_vsql(config_persons, "not (app.p_date_value.value == None)")

def test_datetime_none_false(config_persons):
	check_vsql(config_persons, "app.p_datetime_none.value == None")

def test_datetime_none_true(config_persons):
	check_vsql(config_persons, "not (app.p_datetime_value.value == None)")

def test_color_none_false(config_persons):
	check_vsql(config_persons, "app.p_color_none.value == None")

def test_color_none_true(config_persons):
	check_vsql(config_persons, "not (app.p_color_value.value == None)")

def test_datedelta_none_false(config_persons):
	check_vsql(config_persons, "app.p_datedelta_none.value == None")

def test_datedelta_none_true(config_persons):
	check_vsql(config_persons, "not (app.p_datedelta_value.value == None)")

def test_datetimedelta_none_false(config_persons):
	check_vsql(config_persons, "app.p_datetimedelta_none.value == None")

def test_datetimedelta_none_true(config_persons):
	check_vsql(config_persons, "not (app.p_datetimedelta_value.value == None)")

def test_monthdelta_none_false(config_persons):
	check_vsql(config_persons, "app.p_monthdelta_none.value == None")

def test_monthdelta_none_true(config_persons):
	check_vsql(config_persons, "not (app.p_monthdelta_value.value == None)")

def test_bool_bool_false(config_persons):
	check_vsql(config_persons, "app.p_bool_false.value == False")

def test_bool_bool_true(config_persons):
	check_vsql(config_persons, "app.p_bool_true.value == True")

def test_bool_int_false(config_persons):
	check_vsql(config_persons, "app.p_bool_false.value == 0")

def test_bool_int_true(config_persons):
	check_vsql(config_persons, "app.p_bool_true.value == 1")

def test_int_bool_false(config_persons):
	check_vsql(config_persons, "0 == app.p_bool_false.value")

def test_int_bool_true(config_persons):
	check_vsql(config_persons, "1 == app.p_bool_true.value")

def test_number_bool_false(config_persons):
	check_vsql(config_persons, "0.0 == app.p_bool_false.value")

def test_number_bool_true(config_persons):
	check_vsql(config_persons, "1.0 == app.p_bool_true.value")

def test_number_int_false(config_persons):
	check_vsql(config_persons, "not (42.5 == app.p_int_value.value)")

def test_number_int_true(config_persons):
	check_vsql(config_persons, "1777.0 == app.p_int_value.value")

def test_number_number_false(config_persons):
	check_vsql(config_persons, "not (17.23 == app.p_number_value.value)")

def test_number_number_true(config_persons):
	check_vsql(config_persons, "42.5 == app.p_number_value.value")

def test_str_str_false(config_persons):
	check_vsql(config_persons, "not (app.p_str_none.value == 'gurk')")

def test_str_str_true(config_persons):
	check_vsql(config_persons, "app.p_str_value.value == 'gurk'")

def test_date_date_false(config_persons):
	check_vsql(config_persons, "not (app.p_date_none.value == @(2000-02-29))")

def test_date_date_true(config_persons):
	check_vsql(config_persons, "app.p_date_value.value == @(2000-02-29)")

def test_datetime_datetime_false(config_persons):
	check_vsql(config_persons, "not (app.p_datetime_none.value == @(2000-02-29T12:34:56))")

def test_datetime_datetime_true(config_persons):
	check_vsql(config_persons, "app.p_datetime_value.value == @(2000-02-29T12:34:56)")

def test_datedelta_datedelta_false(config_persons):
	check_vsql(config_persons, "not (app.p_datedelta_none.value == days(12))")

def test_datedelta_datedelta_true(config_persons):
	check_vsql(config_persons, "app.p_datedelta_value.value == days(12)")

def test_color_color_false(config_persons):
	check_vsql(config_persons, "not (app.p_color_none.value == #369c)")

def test_color_color_true(config_persons):
	check_vsql(config_persons, "app.p_color_value.value == #369c")

def test_datetimedelta_datetimedelta_false(config_persons):
	check_vsql(config_persons, "not (app.p_datetimedelta_none.value == timedelta(1, 45296))")

def test_datetimedelta_datetimedelta_true(config_persons):
	check_vsql(config_persons, "app.p_datetimedelta_value.value == timedelta(1, 45296)")

def test_intlist_intlist1(config_persons):
	check_vsql(config_persons, "not ([1] == [2])")

def test_intlist_intlist2(config_persons):
	check_vsql(config_persons, "not ([1] == [1, 2])")

def test_intlist_intlist3(config_persons):
	check_vsql(config_persons, "not ([1, 2] == [1])")

def test_intlist_intlist4(config_persons):
	check_vsql(config_persons, "not ([1, None] == [1])")

def test_intlist_intlist5(config_persons):
	check_vsql(config_persons, "[1, None, 2, None, 3] == [1, None, 2, None, 3]")

def test_intlist_numberlist1(config_persons):
	check_vsql(config_persons, "not ([1] == [1.5])")

def test_intlist_numberlist2(config_persons):
	check_vsql(config_persons, "not ([1] == [1.0, 2.0])")

def test_intlist_numberlist3(config_persons):
	check_vsql(config_persons, "not ([1, 2] == [1.0])")

def test_intlist_numberlist4(config_persons):
	check_vsql(config_persons, "not ([1, None] == [1.0])")

def test_intlist_numberlist5(config_persons):
	check_vsql(config_persons, "[1, None, 2, None, 3] == [1.0, None, 2.0, None, 3.0]")

def test_numberlist_intlist1(config_persons):
	check_vsql(config_persons, "not ([1.5] == [2])")

def test_numberlist_intlist2(config_persons):
	check_vsql(config_persons, "not ([1.0] == [1, 2])")

def test_numberlist_intlist3(config_persons):
	check_vsql(config_persons, "not ([1.0, 2.0] == [1])")

def test_numberlist_intlist4(config_persons):
	check_vsql(config_persons, "not ([1.0, None] == [1])")

def test_numberlist_intlist5(config_persons):
	check_vsql(config_persons, "[1.0, None, 2.0, None, 3.0] == [1, None, 2, None, 3]")

def test_numberlist_numberlist1(config_persons):
	check_vsql(config_persons, "not ([1.5] == [2.5])")

def test_numberlist_numberlist2(config_persons):
	check_vsql(config_persons, "not ([1.0] == [1.0, 2.0])")

def test_numberlist_numberlist3(config_persons):
	check_vsql(config_persons, "not ([1.0, 2.0] == [1.0])")

def test_numberlist_numberlist4(config_persons):
	check_vsql(config_persons, "not ([1.5, None] == [1.5])")

def test_numberlist_numberlist5(config_persons):
	check_vsql(config_persons, "[1.5, None, 2.5, None, 3.5] == [1.5, None, 2.5, None, 3.5]")

def test_strlist_strlist1(config_persons):
	check_vsql(config_persons, "not (['foo'] == ['bar'])")

def test_strlist_strlist2(config_persons):
	check_vsql(config_persons, "not (['foo'] == ['foo', 'bar'])")

def test_strlist_strlist3(config_persons):
	check_vsql(config_persons, "not (['foo', 'bar'] == ['foo'])")

def test_strlist_strlist4(config_persons):
	check_vsql(config_persons, "not (['foo', None] == ['foo'])")

def test_strlist_strlist5(config_persons):
	check_vsql(config_persons, "['foo', None, 'bar', None, 'baz'] == ['foo', None, 'bar', None, 'baz']")

def test_datelist_datelist1(config_persons):
	check_vsql(config_persons, "not ([@(2000-02-29)] == [@(2000-03-01)])")

def test_datelist_datelist2(config_persons):
	check_vsql(config_persons, "not ([@(2000-02-29)] == [@(2000-02-29), @(2000-03-01)])")

def test_datelist_datelist3(config_persons):
	check_vsql(config_persons, "not ([@(2000-02-29), @(2000-03-01)] == [@(2000-02-29)])")

def test_datelist_datelist4(config_persons):
	check_vsql(config_persons, "not ([@(2000-02-29), None] == [@(2000-02-29)])")

def test_datelist_datelist5(config_persons):
	check_vsql(config_persons, "[@(2000-02-29), None, @(2000-03-01), None, @(2000-03-02)] == [@(2000-02-29), None, @(2000-03-01), None, @(2000-03-02)]")

def test_datetimelist_datetimelist1(config_persons):
	check_vsql(config_persons, "not ([@(2000-02-29T12:34:56)] == [@(2000-03-01T12:34:56)])")

def test_datetimelist_datetimelist2(config_persons):
	check_vsql(config_persons, "not ([@(2000-02-29T12:34:56)] == [@(2000-02-29T12:34:56), @(2000-03-01T12:34:56)])")

def test_datetimelist_datetimelist3(config_persons):
	check_vsql(config_persons, "not ([@(2000-02-29T12:34:56), @(2000-03-01T12:34:56)] == [@(2000-02-29T12:34:56)])")

def test_datetimelist_datetimelist4(config_persons):
	check_vsql(config_persons, "not ([@(2000-02-29T12:34:56), None] == [@(2000-02-29T12:34:56)])")

def test_datetimelist_datetimelist5(config_persons):
	check_vsql(config_persons, "[@(2000-02-29T12:34:56), None, @(2000-03-01T12:34:56), None, @(2000-03-02T12:34:56)] == [@(2000-02-29T12:34:56), None, @(2000-03-01T12:34:56), None, @(2000-03-02T12:34:56)]")

def test_intset_intset1(config_persons):
	check_vsql(config_persons, "not ({1} == {2})")

def test_intset_intset2(config_persons):
	check_vsql(config_persons, "not ({1} == {1, 2})")

def test_intset_intset3(config_persons):
	check_vsql(config_persons, "not ({1, 2} == {1})")

def test_intset_intset4(config_persons):
	check_vsql(config_persons, "not ({1, None} == {1})")

def test_intset_intset5(config_persons):
	check_vsql(config_persons, "{1, None, 2, None, 3} == {None, 3, 2, 1, None}")

def test_numberset_numberset1(config_persons):
	check_vsql(config_persons, "not ({1.5} == {2.5})")

def test_numberset_numberset2(config_persons):
	check_vsql(config_persons, "not ({1.5} == {1.5, 2.5})")

def test_numberset_numberset3(config_persons):
	check_vsql(config_persons, "not ({1.5, 2.5} == {1.5})")

def test_numberset_numberset4(config_persons):
	check_vsql(config_persons, "not ({1.5, None} == {1.5})")

def test_numberset_numberset5(config_persons):
	check_vsql(config_persons, "{1.5, None, 2.5, None, 3.5} == {None, 3.5, 2.5, 1.5, None}")

def test_strset_strset1(config_persons):
	check_vsql(config_persons, "not ({1.5} == {2.5})")

def test_strset_strset2(config_persons):
	check_vsql(config_persons, "not ({'foo'} == {'foo', 'bar'})")

def test_strset_strset3(config_persons):
	check_vsql(config_persons, "not ({'foo', 'bar'} == {'foo'})")

def test_strset_strset4(config_persons):
	check_vsql(config_persons, "not ({'foo', None} == {'foo'})")

def test_strset_strset5(config_persons):
	check_vsql(config_persons, "{'foo', None, 'bar', None, 'baz'} == {None, 'baz', 'bar', 'foo', None}")

def test_dateset_dateset1(config_persons):
	check_vsql(config_persons, "not ({@(2000-02-29)} == {@(2000-03-01)})")

def test_dateset_dateset2(config_persons):
	check_vsql(config_persons, "not ({@(2000-02-29)} == {@(2000-02-29), @(2000-03-01)})")

def test_dateset_dateset3(config_persons):
	check_vsql(config_persons, "not ({@(2000-02-29), @(2000-03-01)} == {@(2000-02-29)})")

def test_dateset_dateset4(config_persons):
	check_vsql(config_persons, "not ({@(2000-02-29), None} == {@(2000-02-29)})")

def test_dateset_dateset5(config_persons):
	check_vsql(config_persons, "{@(2000-02-29), None, @(2000-03-01), None, @(2000-03-02)} == {None, @(2000-03-02), @(2000-03-01), @(2000-02-29), None}")

def test_datetimeset_datetimeset1(config_persons):
	check_vsql(config_persons, "not ({@(2000-02-29T12:34:56)} == {@(2000-03-01T12:34:56)})")

def test_datetimeset_datetimeset2(config_persons):
	check_vsql(config_persons, "not ({@(2000-02-29T12:34:56)} == {@(2000-02-29T12:34:56), @(2000-03-01T12:34:56)})")

def test_datetimeset_datetimeset3(config_persons):
	check_vsql(config_persons, "not ({@(2000-02-29T12:34:56), @(2000-03-01T12:34:56)} == {@(2000-02-29T12:34:56)})")

def test_datetimeset_datetimeset4(config_persons):
	check_vsql(config_persons, "not ({@(2000-02-29T12:34:56), None} == {@(2000-02-29T12:34:56)})")

def test_datetimeset_datetimeset5(config_persons):
	check_vsql(config_persons, "{@(2000-02-29T12:34:56), None, @(2000-03-01T12:34:56), None, @(2000-03-02T12:34:56)} == {None, @(2000-03-02T12:34:56), @(2000-03-01T12:34:56), @(2000-02-29T12:34:56), None}")

# FIXME Add tests for mixed type comparisons?
